//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Preferences.cc --- implementation of Preference classes
//
//	DESCRIPTION:
//		This implements all preference classes
//
//	CHANGE HISTORY:
//		UL, 2000-10-09: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers, Mon Oct  9 13:37:19 MET DST 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/preferences/src/Preferences.cc,v 1.3 2000/10/26 13:43:00 ulammers Exp $
//

#include <fstream>
#include <iomanip>
#include "stringutil.h"
#include "ExceptionErrorHandler.h"
#include "errstr.h"
#include "Preferences.h"

namespace {

ostream &indentation(ostream &os, short level)
{
	PROC
	static const unsigned indentWidth = 4;

	os<<setw(level*indentWidth)<<"";

	RETURN os;
} // indentation()

}

//
//	---( PreferenceNode )-----------------------------------------------------
//
PreferenceNode::PreferenceNode(const string &name, short level,
							   const PreferenceValue &value)
	: _group(0), _name(name), _level(level)
{
	PROC

	_value = new PreferenceValue(value);

	ENDPROC
} // PreferenceNode::PreferenceNode()


PreferenceNode::PreferenceNode(const string &name, short level,
							   PreferenceGroup *group)
	: _value(0), _name(name), _level(level)
{
	PROC

	_group = group;

	ENDPROC
} // PreferenceNode::PreferenceNode()


const string &PreferenceNode::name() const
{
	PROC
	RETURN _name;
} // PreferenceNode::name()


bool PreferenceNode::hasValue() const
{
	PROC
	RETURN _value;
} // PreferenceNode::hasValue()


PreferenceValue &PreferenceNode::value() const
{
	PROC

	if (!_value)
		errstr<<"This is a group preference with no value"
			  <<Sas::error("NoValuePreference");

	RETURN *_value;
} // PreferenceNode::value()


bool PreferenceNode::isGroup() const
{
	PROC
	RETURN !hasValue();
} // PreferenceNode::isGroup()


PreferenceGroup *PreferenceNode::group() const
{
	PROC

	if (!_group)
		errstr<<"This is a value preference with no sub-groups"
			  <<Sas::error("NoGroupPreference");

	RETURN _group;
} // PreferenceNode::group()


void PreferenceNode::print(ostream &os) const
{
	PROC

	indentation(os, _level)<<'"'<<name()<<'"';
	if (hasValue())
		os<<'\t'<<value()<<endl;
	else {
		os<<" {\n";
		group()->print(os);
		indentation(os, _level)<<"}\n";
	}

	ENDPROC		
} // PreferenceNode::print()


PreferenceNode::~PreferenceNode()
{
	PROC

	delete _group;
	delete _value;

	ENDPROC
} // PreferenceNode::~PreferenceNode


//
//	---( PreferenceGroup )----------------------------------------------------
//
PreferenceGroup::PreferenceGroup(const string &name, short level)
	: _name(name), _level(level)
{
	PROC
	ENDPROC
} // PreferenceGroup::PreferenceGroup()


void PreferenceGroup::addPreference(const string &name,
									const PreferenceValue &value)
{
	PROC

	this->insert(PreferenceMap::value_type(name,
										   new PreferenceNode(name, _level+1,
															  value)));

	ENDPROC
} // PreferenceGroup::addPreference()


PreferenceGroup *PreferenceGroup::addPreferenceGroup(const string &name)
{
	PROC

	PreferenceGroup *pg = new PreferenceGroup(name, _level+1);
	this->insert(PreferenceMap::value_type(name,
								   new PreferenceNode(name, _level+1, pg)));

	RETURN pg;
} // PreferenceGroup::addPreference()


vector<PrefMapIterator>	PreferenceGroup::valuePreferences() const
{
	PROC
	vector<PrefMapIterator> v;
	for (PrefMapIterator i = begin(); i != end(); ++i)
		if (i->second->hasValue())
			v.push_back(i);

	RETURN v;
} // PreferenceGroup::valuePreferences()


vector<PrefMapIterator>	PreferenceGroup::groupPreferences() const
{
	PROC
	vector<PrefMapIterator> v;
	for (PrefMapIterator i = begin(); i != end(); ++i)
		if (i->second->isGroup())
			v.push_back(i);

	RETURN v;
} // PreferenceGroup::groupPreferences()


bool PreferenceGroup::hasPreference(const string &prefPath) const
{
	PROC

	ExceptionErrorHandler local;
	ErrorHandler *oldHandler = errHandler.handler(&local);
	bool res = true;
	try {
		preference(prefPath);
	} catch (...) {
		res = false;
	}

	errHandler.handler(oldHandler);

	RETURN res;
} // PreferenceGroup::hasPreference()


PreferenceNode *PreferenceGroup::preference(const string &prefPath) const
{
	PROC

	string::size_type p = prefPath.find_first_of(".", 0);
	if (p != string::npos) {
		string key(prefPath.substr(0, p));
		PrefMapIterator i = find(key);
		if (i != end()) {
			const PreferenceNode *n = i->second;
			if (n->isGroup()) {
				ALTRETURNV(n->group()->preference(prefPath.substr(p+1)));
			} else
				errstr<<"`"<<prefPath<<"' is a value preference"
					  <<Sas::error("IllegalPreferenceSpecifier");
		} else
			errstr<<"There is no preference `"<<key<<"'"
				  <<Sas::error("NoSuchPreference");
	}

	PrefMapIterator i = find(prefPath);
	if (i == end())
		errstr<<"There is no preference `"<<prefPath<<"'"
			  <<Sas::error("NoSuchPreference");

	RETURN i->second;
} // PreferenceGroup::preference()


void PreferenceGroup::print(ostream &os) const
{
	PROC

	for (PrefMapIterator i = begin(); i!=end(); (i++)->second->print(os));

	ENDPROC
} // PreferenceGroup::print()
		

PreferenceGroup::~PreferenceGroup()
{
	PROC

	for (PrefMapIterator i = begin(); i != end(); ++i)
		delete i->second;

	ENDPROC
} // PreferenceGroup::~PreferenceGroup()


//
//	---( PreferenceInStream )------------------------------------------------
//
inline void PreferenceInStream::skipRestOfLine()
{
	PROC
	readline();
	ENDPROC
} // skipRestOfLine


template<class T>
inline void PreferenceInStream::operator>>(T &to)
{
	PROC
	string line(readline());
	strstream str; str<<line;
	str>>to;
} // operator>>


inline string PreferenceInStream::readline()
{
	PROC
	char s[256];
	getline(s, 256, '\n');
	string line(trim(s, true));
	while ((line.length() == 0 ||
		   (line.length()>=2 && line[0] == '/' && line[1] == '/'))
			&& *this) {
		getline(s, 256, '\n');
		line = trim(s, true);
	}
	RETURN line;
} // PreferenceInStream::readline()


void PreferenceInStream::readGroup(PreferenceGroup &group, bool matchName)
{
	PROC

	if (matchName && group.name().empty()) {
		errstr<<"Cannot match empty group name in reading preference group"
			  <<Sas::warning("EmptyPreferenceGroupName");
		ALTRETURN;
	}

	string line(readline());
	while (!line.empty()) {
		vector<string> v(quotedBreakup(line, " \t", "\""));
		if (v.size() == 1 && v[0] == "}")
			ALTRETURN;	//	end of group
		if (v.size() != 2)
			errstr<<"Cannot interpret line `"<<line<<"' in preference file `"
				  <<_name<<"'"<<Sas::error("InvalidPreferenceFile");

		if (v[1] != "{") {
			if (!matchName)
				//	normal reference
				group.addPreference(unquote(v[0]),
									PreferenceValue(unquote(v[1])));
		} else {
			//	a group
			string groupName(unquote(v[0]));
			if (matchName)
				if (groupName == group.name())
					readGroup(group);
				else {
					//	read entire group into dummy PreferenceGroup
					PreferenceGroup pg;
					readGroup(pg);
				}
			else {
				//	beginning of a group
				PreferenceGroup *gp = group.addPreferenceGroup(unquote(v[0]));
				readGroup(*gp);
			}
		}

		line = readline();
	}

	ENDPROC
} // PreferenceInStream::readGroup()


PreferenceInStream &operator>>(PreferenceInStream &is, PreferenceGroup &group)
{
	PROC
	is.readGroup(group, is._matchGroupName && !group.name().empty());
	RETURN is;
} // operator>>


//
//	---( PreferenceOutStream )------------------------------------------------
//
PreferenceOutStream &operator<<(PreferenceOutStream &os,
								const PreferenceGroup &group)
{
	PROC
	os<<"//\n"
		"//\tThis is a SAS preference file\n"
		"//\tDo not edit manually unless you know exactly what you are"
		" doing!\n//\n";
	group.print(os);
	RETURN os;

} // operator<<

