//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Preferences.h --- classes supporting the concept of object preferences
//
//	DESCRIPTION:
//		A set of classes is declared which support the concept of object
//		preferences. A preference is a key-value pair where the key can be
//		thought of as a name of an object's resource. Preferences can be
//		grouped	and possess a hierachical structure, i.e., a list of
//		preferences can	contain groups of other preferences. The level of
//		nesting is a priori	unlimited. A preference object should be
//		persistent, hence, outlive the livetime of the associated described
//		main object.
//
//		In the context of the SAS:
//			* Objects with preferences are primarily tasks
//			* Persistence should be implemented via preference files
//			* All interactive GUI tasks should have preferences
//			* The default values of a task's parameters should be preferences
//			* Preference files should be user specific
//			* There should be a standard mechanism for the user to
//			  create/edit/modify/delete task preferences (preference editor)
//
//		Example: Possible hierachy of preferences of task 'xmmselect':
//
//			centerpaneheight	6
//			quicklook           yes
//			Expressions {
//				"MOS Selections" {
//					"bad pixels"	"FLAG&0x133300==0"
//					Circle		    "(DETX,DETY) in circle(100,10,10)"
//				}
//				"PN Selections" {
//					"bad pixels"	"FLAG&0x133331==0"
//					Fred			"(NAME=="FRED")"
//				}
//			}
//
//	CHANGE HISTORY:
//		UL, 2000-10-09: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers, Mon Oct  9 13:37:19 MET DST 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/preferences/src/Preferences.h,v 1.3 2000/10/26 13:43:00 ulammers Exp $
//
#ifndef PREFERENCES_H
#define PREFERENCES_H

#include <string>
#include <iostream>
#include <fstream>
#include <map>

#include "Tracer.h"
#include "stringutil.h"
#include "errstr.h"

//
//	---( PreferenceValue )------------------------------------------------
//
class PreferenceValue : public string {
public:
	explicit PreferenceValue(const string &s) : string(s) { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Construct PreferenceValue object
		//


	bool asBool() const {	PROC
							string s(uppercase(*this));	
							if (s=="TRUE" || s=="YES" || s=="+")
								ALTRETURNV(true);
							if (s=="FALSE" || s=="NO" || s=="-")
								ALTRETURNV(false);
							errstr<<"`"<<*this<<"' is not a valid boolean"
								  <<Sas::error("NotBoolean");
							RETURN false;
						}
		//
		//	PURPOSE:
		//		Access preference value as a boolan.
		//


	template<class T>
	T asNumber() const { PROC RETURN toNumber<T>(*this); }
		//
		//	PURPOSE:
		//		Access preference value as a number.
		//


	string asString() const { PROC RETURN *this; }
		//
		//	PURPOSE:
		//		Access preference value as a string.
		//


	~PreferenceValue() { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Destroy PreferenceValue object.
		//
};

//
//	print preference value to an ostream
//
inline ostream &operator<<(ostream &os, const PreferenceValue &v)
{
	PROC
	RETURN os<<'"'<<(string)v<<'"';
} // operator<<


//
//	---( PreferenceNode )----------------------------------------------------
//
class PreferenceGroup;	//	forward declaration

//
//	A PreferenceNode has a name and has either a PreferenceValue or
//	points to a preference group.
//
class PreferenceNode {
public:
	const string &name() const;
		//
		//	PURPOSE:
		//		Returns this node's name.
		//


	bool hasValue() const;
	PreferenceValue &value() const;
		//
		//	PURPOSE:
		//		Inquire whether this node holds a value/Access this node's
		//		value; attempts to access the value of group nodes will
		//		lead to an error condition.
		//


	bool isGroup() const;
	PreferenceGroup *group() const;
		//
		//	PURPOSE:
		//		Inquire whether this node points to a preference group/Access
		//		preference group this node points to; attempts to access
		//		preference group pointer for value nodes will lead to
		//		an error condition.
		//


	void print(ostream &os) const;
		//
		//	PURPOSE:
		//		Print node on ostream.
		//


	~PreferenceNode();
		//
		//	PURPOSE:
		//		Destroy object.
		//

private:
	//	only PreferenceGroup objects can instaniate nodes
	friend class PreferenceGroup;
	explicit PreferenceNode(const string &name, short level,
							const PreferenceValue &value);
	explicit PreferenceNode(const string &name, short level,
							PreferenceGroup *group);

private:
	PreferenceValue	*_value;
	PreferenceGroup	*_group;
	string			_name;
	short			_level;
};


//
//	---( PreferenceMap/PreferenceGroup )-------------------------------------
//
typedef map<string, PreferenceNode *, less<string> > PreferenceMap;
typedef PreferenceMap::const_iterator PrefMapIterator;

class PreferenceGroup : public PreferenceMap {
public:
	explicit PreferenceGroup(const string &name = "", short level = -1);
		//
		//	PURPOSE:
		//		Create PreferenceGroup object
		//
		//	ARGUMENTS:
		//		name	: name of this group
		//		level	: level in hierachy (-1 == root level)
		//


	const string &name() const { PROC RETURN _name; }
		//
		//	PURPOSE:
		//		Get/set groups name.
		//


	void addPreference(const string &name, const PreferenceValue &value);
		//
		//	PURPOSE:
		//		Add a value preference to this group.
		//
		//	ARGUMENTS:
		//		name	: name of the value preference to add
		//		value	: value of the newly added preference
		//


	PreferenceGroup *addPreferenceGroup(const string &name);
		//
		//	PURPOSE:
		//		Insert a new empty preference group into the current group;
		//		the level of the newly added group will be one larger than
		//		that of the parent group.
		//
		//	ARGUMENTS:
		//		name	: name of the new group
		//


	bool hasPreference(const string &prefPath) const;
		//
		//	PURPOSE:
		//		Test whether a preference with a given qualified name exists in
		//		the hierachy beneath the current preference group. The
		//		qualified name is a dot-separated list of preference group
		//		names, e.g.
		//			"xmmselect.selections.MOSselections.badPixels"
		//		The	last component can either be a value or a group preference
		//
		//	RETURNS:
		//		true if the names preference does exist in the hierachy
		//		beneath the current group, or false, if:
		//			* the named preference does not exist
		//			* the name of a value preference is used where a group
		//			  preference name was expected
		//	


	PreferenceNode *preference(const string &prefPath) const;
		//
		//	PURPOSE:
		//		Access a preference in the preference hierachy beneath the
		//		current group through its qualified name
		//
		//	RETURNS:
		//		Pointer to PreferenceNode; an error condition will occur
		//		for all <prefPath> values for which hasPreference() returns
		//		false
		//


	vector<PrefMapIterator>	valuePreferences() const;
	vector<PrefMapIterator> groupPreferences() const;
		//
		//	PURPOSE:
		//		Returns vector with iterators to all value/group preferences
		//		in the group.
		//


	void print(ostream &os) const;
		//
		//	PURPOSE:
		//		Print this group to an ostream.
		//


	~PreferenceGroup();
		//
		//	PURPOSE:
		//		Destroy this preference group.
		//

private:
	string	_name;
	short	_level;
};


//
// ---( PreferenceOutStream / PreferenceInStream )----------------------------
//
class PreferenceOutStream : public ofstream {
public:
	explicit PreferenceOutStream(const string &name)
		: ofstream(name.c_str()) { PROC ENDPROC }

	~PreferenceOutStream() { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Construct/destroy PreferenceOutStream
		//
		//	ARGUMENTS:
		//		name	: name of preference file to write to
		//
};

PreferenceOutStream &operator<<(PreferenceOutStream &os,
								const PreferenceGroup &group);
	//
	//	PURPOSE:
	//		Print preference group to a PreferenceOutStream
	//


class PreferenceInStream : public ifstream {
public:
	explicit PreferenceInStream(const string &name, bool matchGroupName = true)
		: ifstream(name.c_str()), _name(name), _matchGroupName(matchGroupName)
	{ PROC ENDPROC }

	void matchGroupName(bool match = true) { PROC _matchGroupName = match;
											 ENDPROC }

	~PreferenceInStream() { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Construct/destroy PreferenceInStream
		//
		//	ARGUMENTS:
		//		name	: name of preference file to read from
		//

private:
	friend PreferenceInStream &operator>>(PreferenceInStream &is,
										  PreferenceGroup &group);

	void skipRestOfLine();
	template<class T>
	void operator>>(T &to);
	string readline();
	void readGroup(PreferenceGroup &group, bool matchName = false);

	string	_name;
	bool	_matchGroupName;
};

PreferenceInStream &operator>>(PreferenceInStream &is, PreferenceGroup &group);
	//
	//	PURPOSE:
	//		Read contents of preference file into PreferenceGroup object
	//		Syntax rules are:
	//			* lines starting with '//' are ignored
	//			* all other lines must contain
	//				1) PrefKey PrefVal
	//				2) PrefKey "{"
	//				3) "}"
	//			* PrefKey/PrefVal can optionally be enclosed in "" to protect
	//			  white space characters
	//			* white space outside "" is ignored everywhere
	//			* 1) is a preference value defintion
	//			* 2) is the beginning of a preference group
	//			* 3) is the end of a preference group
	//

#endif
