//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	sasprefgen.cc --- create SAS-wide preference file
//
//	DESCRIPTION:
//		The task creates a SAS-wide preference file with the name
//		TaskPreferenceGroup::sasPrefFileName (see TaskPreferences.h) in the
//		user's home directory. This is the result of a compilation of
//		all task-specific preferences file which end in .pref and exist
//		along the SAS_DIR:SAS_PATH path.
//
//		If a SAS-preference file exists already it will be renamed by
//		adding the suffix .old before the new file is written.
//
//	CHANGE HISTORY:
//		UL, 2000-26-10: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers, Thu Oct 26 13:30:45 MET DST 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/preferences/src/sasprefgen.cc,v 1.2 2000/10/27 14:38:56 ulammers Exp $
//

#include <cstdlib>

#include "Preferences.h"
#include "TaskPreferences.h"
#include "errstr.h"
#include "Directory.h"
#include "FileLocator.h"

#include "sasprefgen.h"

void sasprefgen()
{
	char *sasDir = getenv("SAS_DIR");
	if (!sasDir)
		errstr<<"environment variable SAS_DIR not set"
			  <<Sas::error("SAS_DIRnotSet");
	string sasPath(sasDir);
	char *sasPathp = getenv("SAS_PATH");
	if (sasPathp)
		sasPath = string(sasPathp) + ':' + sasPath;

	vector<string> dirs(breakup(sasPath, ":"));
	PreferenceGroup sasPrefs;
	for (unsigned n=0; n<dirs.size(); ++n) {
		errstr<<"Scanning directory "<<dirs[n]<<" for preference files ..."
			  <<Sas::message(Msg::App, Msg::Always);
		Directory d(dirs[n], "*.pref", true);
		typedef Directory::const_iterator	FileListIter;
		for (FileListIter i = d.begin(); i!=d.end(); ++i) {
			string taskName((*i)->basename());
			taskName.replace(taskName.find(".pref"), 5, "");
			//
			//	we have the task name - check if there is already
			//	a preference group; if yes, we skip it, if no we add one
			//
			errstr<<"Found "<<(*i)->name()
				  <<Sas::message(Msg::App, Msg::Always);
			if (sasPrefs.hasPreference(taskName))
				errstr<<"There is already a preference group for task `"
					  <<taskName<<"' - "<<(*i)->name()<<" will be ignored - "
						"you may want to reorder the directories in SAS_PATH"
					  <<Sas::warning("PreferenceGroupIgnored");
			else {
				PreferenceGroup *newGrp =
										sasPrefs.addPreferenceGroup(taskName);
				PreferenceInStream pin((*i)->name(), false);
				pin>>*newGrp;
			}
		}
	}

	string home(getenv("HOME")),
		   prefFile = home + '/' +string(TaskPreferenceGroup::sasPrefFileName);
	FileLocator floc(home);
	if (!floc.locate(TaskPreferenceGroup::sasPrefFileName, "", false,
															false).empty()) {
		string cmd = string("mv ") + prefFile + ' ' + prefFile + ".old";
		system(cmd.c_str());
		errstr<<prefFile<<" exists already - moving to "<<prefFile<<".old and "
				"rewriting"<<Sas::warning("PreferenceFileExists");
	}
	PreferenceOutStream pout(prefFile);
	pout<<sasPrefs;

	errstr<<"SAS-wide preferences written to `"<<prefFile<<"'."
		  <<Sas::message(Msg::App, Msg::Always);

} // sasprefgen()
